#ifndef OSL_CATEGORY_MAJOR_PIECE_H
#define OSL_CATEGORY_MAJOR_PIECE_H

#include "osl/category/categoryEnv.h"
#include "osl/category/categoryTraits.h"
#include "osl/category/probTable.h"
#include "osl/container/moveStack.h"
#include "osl/move_action/store.h"
#include "osl/move_generator/safeDropMajorPiece.h"
#include "osl/move_classifier/directCheck.h"

namespace osl
{
  namespace category
  {
    struct SafeDropMajorPiece : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 321
      static const char *getName() { return "SafeDropMajorPiece"; }
      static void generate(const CategoryEnv& env, MoveVector& out) {
	move_action::Store store(out);
	if (env.state->getTurn() == BLACK)
	  move_generator::SafeDropMajorPiece<BLACK>::generate(*env.state, store);
	else
	  move_generator::SafeDropMajorPiece<WHITE>::generate(*env.state, store);
      }
    };

    struct PrepareRookPromotion : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 300
      static const char *getName() { return "PrepareRookPromotion"; }
      static size_t getIndex(const CategoryEnv&, Move)
      {
	return 0;
      }
      static std::pair<bool,size_t>
      getIndexIf(const CategoryEnv& e, Move m, int /*piece_value*/)
      {
	if (!isInCategory(e, m))
	  return std::make_pair(false, 0u);
	return std::make_pair(true, getIndex(e,m));
      }

      static bool isInCategory(const CategoryEnv& e, Move m)
      {
	Player p = e.state->getTurn();
	if (m.ptype() == ROOK && !m.to().canPromote(p)
	    && (m.from().y() == m.to().y() ||
		isPiece(m.capturePtype()))
	    && !e.state->hasEffectBy(alt(p), m.to()))
	{
	  for (Position pos = Board_Table.nextPosition(p, m.to(), U);
	       pos.isOnBoard();
	       pos = Board_Table.nextPosition(p, pos, U))
	  {
	    Piece piece = e.state->getPieceOnBoard(pos);
	    if (!piece.isEmpty() && piece.owner() == p)
	      return false;

	    if (pos.canPromote(p) &&
		! e.state->hasEffectBy(alt(p), pos))
	      return true;

	    if (!piece.isEmpty())
	      return false;
	  }
	}
	return false;
      }
    };

    // ʤƤɤ
    struct RookAddPawnSupport : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 400
      static const char *getName() { return "RookAddPawnSupport"; }
      static size_t getIndex(const CategoryEnv&, Move)
      {
	return 0;
      }
      static std::pair<bool,size_t>
      getIndexIf(const CategoryEnv& e, Move m, int /*piece_value*/)
      {
	if (!isInCategory(e, m))
	  return std::make_pair(false, 0u);
	return std::make_pair(true, getIndex(e,m));
      }

      static bool isInCategory(const CategoryEnv& e, Move m)
      {
	Player p = e.state->getTurn();
	if (m.ptype() == ROOK
	    && (m.from().y() == m.to().y() ||
		isPiece(m.capturePtype()))
	    && !e.state->hasEffectBy(alt(p), m.to())
	    && e.state->isPawnMaskSet(p, m.to().x()))
	{
	  for (Position pos = Board_Table.nextPosition(p, m.to(), U);
	       pos.isOnBoard();
	       pos = Board_Table.nextPosition(p, pos, U))
	  {
	    Piece piece = e.state->getPieceOnBoard(pos);
	    if (!piece.isEmpty())
	    {
	      if (piece.ptype() == PAWN && piece.owner() == p)
	      {
		const Position target = Board_Table.nextPosition(p, pos, U);
		if (target.isOnBoard() && target.canPromote(p) &&
		    e.state->countEffect(BLACK, target) ==
		    e.state->countEffect(WHITE, target))
		{
		  return true;
		}
	      }
	      return false;
	    } 
	  }
	}
	return false;
      }
    };

    /**
     *  ˤ
     *  դ˼ʬζˤʤäƤ뤫ɤϸΤƤʤ
     */
    struct OpenAttack : FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 300
      static const char *getName() { return "OpenAttack"; }
      template<Ptype type> static bool isInCategory(const CategoryEnv& e, Move m, Player p)
      {
	if (e.state->hasEffectByPtype<type>(p, m.from())
	    && unpromote(m.ptype()) != type)
	{
	  for (int i = PtypeTraits<type>::indexMin;
	       i < PtypeTraits<type>::indexLimit; i++)
	  {
	    const Piece piece = e.state->getPieceOf(i);
	    if (!piece.isOnBoardByOwner(p)
		|| e.state->hasEffectBy(alt(p), piece.position()))
	      continue;
	    if (e.state->hasEffectTo(piece, m.from()))
	    {
	      Offset move_offset = Board_Table.getShortOffsetNotKnight(Offset32(m.to(), m.from()));
	      Offset open_offset = Board_Table.getShortOffsetNotKnight(Offset32(m.from(), piece.position()));
	      Offset open_offset_reverse = Board_Table.getShortOffsetNotKnight(Offset32(piece.position(), m.from()));

	      assert(!open_offset.zero());
	      if (move_offset == open_offset ||
		  move_offset == open_offset_reverse)
		continue;

	      for (Position pos = m.from() + open_offset;
		   pos.isOnBoard();
		   pos += open_offset)
	      {
		Piece piece = e.state->getPieceAt(pos);
		if (!piece.isEmpty())
		{
		  if (piece.owner() == p || !isMajor(piece.ptype()))
		    break;

		  if (!e.state->hasEffectBy(alt(p), pos) /* &&
		      (unpromote(piece.ptype()) != type ||
		       (p == BLACK &&
			move_classifier::DirectCheck<BLACK>::isMember(*e.state,
								      m.ptype(),
								      m.to()) ||
			(p == WHITE &&
			 move_classifier::DirectCheck<WHITE>::isMember(*e.state,
								       m.ptype(),
								       m.to())))) */)
		    return true;
		  else
		    return false;
		}
	      }
	    }
	  }
	}
	return false;
      }

      static bool isInCategory(const CategoryEnv& e, Move m)
      {
	const Player p = m.player();
	if (m.isDrop() || !e.state->hasEffectBy(p, m.from()))
	  return false;
	bool result = isInCategory<ROOK>(e, m, p);
	if (result)
	  return result;
	return isInCategory<BISHOP>(e, m, p);
      }
    };

    struct RookAddLanceSupport : FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 400
      static const char *getName() { return "RookAddLanceSupport"; }
      static bool isInCategory(const CategoryEnv& e, Move m)
      {
	if (m.ptype() != ROOK || m.from().x() == m.to().x())
	  return false;
	const Player p = m.player();
	for (int i = PtypeTraits<LANCE>::indexMin;
	     i < PtypeTraits<LANCE>::indexLimit; i++)
	{
	  const Piece piece = e.state->getPieceOf(i);
	  if (piece.isOnBoardByOwner(p) &&
	      piece.position().x() == m.to().x() &&
	      ((p == BLACK && piece.position().y() - m.to().y() < 0)
	       || (p == WHITE && piece.position().y() - m.to().y() > 0)))
	    return true;
	}
	return false;
      }
    };
  }
}

#endif // OSL_CATEGORY_MAJOR_PIECE_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
