/* fixedCapacityVector.h
 */
#ifndef _FIXED_CAPACITY_VECTOR_H
#define _FIXED_CAPACITY_VECTOR_H

#include "osl/misc/carray.h"
#include <algorithm>
#include <cstddef>
#include <cassert>
namespace osl
{
  namespace misc
  {
  template <typename T, size_t Capacity>
  class FixedCapacityVector
  {
  protected:
    typedef CArray<T, Capacity> array_t;
    CArray<int, (sizeof(T[Capacity])+sizeof(int)-1)/sizeof(int)> relements;
    size_t index;
  private:
    const array_t &elements() const{
      return *reinterpret_cast<const array_t*>(&relements);
    }
    array_t &elements(){
      return *reinterpret_cast<array_t*>(&relements);
    }
  public:
    typedef typename array_t::iterator iterator;
    typedef typename array_t::const_iterator const_iterator;
    
    FixedCapacityVector() : index(0) {}
    ~FixedCapacityVector() 
    {
      std::_Destroy(begin(),end());
    }
    FixedCapacityVector(FixedCapacityVector const& rhs){
      index=rhs.size();
      std::uninitialized_copy(rhs.begin(),rhs.end(),begin());
    }
    FixedCapacityVector& operator=(FixedCapacityVector const& rhs){
      if(size()>rhs.size()){
	iterator it=std::copy(rhs.begin(),rhs.end(),begin());
	std::_Destroy(it,end());
      }
      else{
	iterator it=std::copy(&(rhs.elements()[0]),
			      &(rhs.elements()[size()]),begin());
	std::uninitialized_copy(&(rhs.elements()[size()]),
				&(rhs.elements()[rhs.size()]),it);
      }
      index=rhs.size();
      return *this;
    }

    T& operator[] (size_t i)
    {
      assert(i < size());
      return elements()[i];
    }

    iterator begin() {  return &elements()[0]; }
    iterator end() { return &elements()[0]+index; }

    void push_back(const T& e)
    {
      assert(index < Capacity);
      std::_Construct(&elements()[index++],e);
    }
    template <class RangeIterator>
    void push_back(const RangeIterator& first, const RangeIterator& last)
    {
      iterator insert_point = end();
      index += (last - first);
      assert(index <= Capacity);
      std::uninitialized_copy(first, last, insert_point);
    }
    void pop_back() { 
      std::_Destroy(&elements()[index]);
      --index; 
    }
    // ʬdestructorƤ
    void clear() { 
      std::_Destroy(begin(),end());
      index = 0; 
    }

    void erase(const T& e)
    {
      const iterator new_end = std::remove(begin(), end(), e);
      std::_Destroy(new_end,end());
      index -= (end() - new_end);
    }

    /** ʣǤ */
    void unique()
    {
      std::sort(begin(),end());
      iterator last = std::unique(begin(), end());
      std::_Destroy(last,end());
      index = (last - begin());
    }

    size_t size() const { return index; }
    bool empty() const { return size() == 0; }
    size_t capacity() const { return Capacity; }

    T const& operator[] (size_t i) const
    {
      assert(i < size());
      return elements()[i];
    }
    const_iterator begin() const { return &elements()[0]; }
    const_iterator end()   const { return &elements()[0]+index; }
    
    bool isMember(const T& e, const_iterator first, const_iterator last) const
    {
      return std::find(first, last, e) != last;
    }
    bool isMember(const T& e) const
    {
      return isMember(e, begin(), end());
    }
  };
  } // namespace misc
  using misc::FixedCapacityVector;
} // namespace osl


#endif /* _FIXED_CAPACITY_VECTOR_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
