#include "osl/brinkmate/searcher.h"
#include "osl/move_generator/legalMoves.h"
#include "osl/move_classifier/check.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/move_order/captureEstimation.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/effect_util/neighboring8Direct.h"
#include "osl/effect_util/additionalEffect.h"
#include "osl/container/moveVector.h"
#include "osl/record/csa.h"
#include "osl/misc/alarm.h"
#include "osl/centering3x3.h"
#include "osl/neighboring8.h"
#include "osl/boardTable.h"
#include <iostream>
#include <string>

static const int report_do_undo = 2;
static const int max_threshold = 1600;

osl::brinkmate::
Searcher::Searcher(Player attack, size_t capacity, int verbose)
  : table(attack, capacity), 
    checkmator(CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT*64),
    stop_flag(0),
    count(0), max_depth(0), verbose(verbose)
{
  state.init(checkmator);
}

osl::brinkmate::
Searcher::~Searcher()
{
}

const osl::Move osl::brinkmate::
Searcher::search(HashEffectState& state, int depth)
{
  max_depth = depth;
  int t = 100;
  int c = 200;
  int d = 0;
  int r = 100;
  BrinkmateResult last_result = UNKNOWN;
  int solution_depth = -1;
  try
  {
    while (true)
    {
      if (verbose)
	std::cerr << "threatmate " << t 
		  << ", checkmate  " << c
		  << ", defense check  " << d
		  << ", checkmate-1 " << r << "\n";
      for (int i=1; i<=depth; i++)
      {
	if (verbose)
	  std::cerr << "depth " << i 
		    << ", count " << count 
		    << ", table " << table.size() << "\n";
	last_result = search(state, i, t, c, d, r);
	if (solution_depth <= i)
	{
	  showBestMoves();
	  if (isAttackSuccess(last_result))
	  {
	    solution_depth = i;
	    if (verbose)
	    {
	      std::cerr << "success!" << " count " << count 
			<< " / " << table.size()
			<< ", result " << last_result << "\n";
	    }
	    break;
	  }
	  else
	  {
	    if (solution_depth >= 0)
	    {
	      solution_depth = -1;
	      if (verbose)
	      {
		std::cerr << "escape found!" << " count " << count 
			  << " / " << table.size()
			  << ", result " << last_result << "\n";
	      }
	      // not break to go deeper
	    }
	  }
	}
      }
      if (isAttackSuccess(last_result))
      {
	assert(solution_depth >= 0);
	r*=2;
	d+=1;
	if (r > max_threshold)
	  break;
      }
      else
      {
	assert(solution_depth < 0);
	t*=2;
	c*=2;
	if (t > max_threshold)
	  break;
      }
    }
  }
  catch(TableFull&)
  {
    if (verbose)
      std::cerr << "table full in brinkmate\n";
  }
  catch(std::bad_alloc&)
  {
    std::cerr << "bad alloc in brinkmate\n";
    this->state.setEmergency();
  }
  if (solution_depth < 0)
    return Move::INVALID();

  const BrinkmateRecord *record = table.find(state.getHash());
  assert(record);
  return record->best_move;
}

osl::brinkmate::BrinkmateResult osl::brinkmate::
Searcher::search(HashEffectState& state,
		 int depth, int tnode, int cnode,
		 int dcm, int rnode)
{
  assert(this->state.depth() == 0);
  assert(this->state.defenseCheckCount() == 0);

  if (this->state.state() != state)
  {
    this->state.init(state);
    killer_moves.clear();
  }
  root_conf.depth = depth;
  root_conf.threatmate_node = tnode;
  root_conf.checkmate_node = cnode;
  root_conf.defense_checkmove_count = dcm;
  root_conf.defense_checkmate_node = rnode;

  // test root threatmate by defender
  if (! EffectUtil::isKingInCheck(state.getTurn(), state))
  {
    const bool root_threatmate
      = this->state.isThreatmateState(root_conf.threatmate_node);
    if (verbose >= report_do_undo  && root_threatmate)
      std::cerr << "root threatmate\n";
  }
  
  // attack
  const BrinkmateResult result = attack();
  assert(isSearchResult(result));

  assert(this->state.depth() == 0);
  assert(this->state.defenseCheckCount() == 0);

  return result;
}

bool osl::brinkmate::
Searcher::threatmateCandidate(const BrinkmateState& state, Move move,
			      Position king)
{
  const Position to = move.to();
  const Move last_move = state.lastMove();
  if (last_move.to() != to)	// allow takeback
    return true;
  if (Neighboring8::isNeighboring8(to, king))
    return true;
  if (Neighboring8Direct::hasEffect
      (state.state(), move.ptypeO(), to, king))
    return true;
  if (last_move.isNormal()
      && Ptype_Table.hasLongMove(last_move.ptype()))
  {
    const Piece last_piece = state.state().getPieceOnBoard(last_move.to());
    if (state.state().hasEffectByPiece(last_piece, to))
      return true;
  }
  // lance as additional effect
  if ((move.ptype() == LANCE)
      && (abs(king.x() - to.x()) <= 1)
      && (move.isDrop() || (move.capturePtype() != PTYPE_EMPTY)))
  {
    const Offset front = Board_Table.getOffset(move.player(), U);
    Position rook_position = to + front;
    Piece rook;
    while ((rook = state.state().getPieceAt(rook_position)) == Piece::EMPTY())
      rook_position += front;
    if (rook.isPiece() && (unpromote(rook.ptype())==ROOK)
	&& (rook.owner() == move.player())
	&& ((to.y() - king.y())*front.dy() > 0))
    {
      return true;
    }
  }
  
  // last entry
  if ((move.capturePtype() == PTYPE_EMPTY)
      || (unpromote(move.capturePtype()) == PAWN))
    return false;
  return true;
}

void osl::brinkmate::
Searcher::setUpAttack(BrinkmateRecord *record)
{
  const int depth_left = depthLeft();
  const bool in_check
    = EffectUtil::isKingInCheck(state.turn(), state.state());
  if ((record->conf.threatmate_node < root_conf.threatmate_node)
      || (record->conf.checkmate_node < root_conf.checkmate_node)
      || (record->conf.defense_checkmove_count 
	  < root_conf.defense_checkmove_count)
      || (record->conf.defense_checkmate_node 
	  < root_conf.defense_checkmate_node))
  {
    record->conf.depth = depth_left; // may need to decrease
    if (! record->moves.empty())
      record->result = WORST_ATTACK; // more accurate?
  }
  else
  {
    record->conf.depth = std::max(record->conf.depth, depth_left);
  }
  if (record->conf.threatmate_node < root_conf.threatmate_node)
  {
    record->conf.threatmate_node = root_conf.threatmate_node;
    record->result = WORST_ATTACK;
    
    MoveVector moves;
    LegalMoves::generate(state.state(), moves);
    if (moves.empty())
    {
      record->result = NOT_BRINKMATE;
      record->conf.threatmate_node = max_threshold;
      record->conf.checkmate_node = max_threshold;
    }

    const Position raw_king
      = state.state().getKingPosition(alt(state.turn()));
    const ThreatmatePool::vector_t& candidates = pool.find(raw_king);

    const Position king = Centering3x3::adjustCenter(raw_king);
    for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
    {
      if (std::find(record->moves.begin(), record->moves.end(), *p) 
	  != record->moves.end())
	continue;

      using namespace osl::move_classifier;
      if (PlayerMoveAdaptor<Check>::isMember(state.state(), *p))
      {
	// check
	record->moves.push_back(*p);
	continue;
      }

      if ((state.depth() > 0) && (! in_check)
	  && (std::find(candidates.begin(), candidates.end(), *p)
	      == candidates.end()) // consider threatmate move in other node
	  && (! threatmateCandidate(state, *p, king)))
	continue;
      
      // threat
      state.push(*p, false);

      const ProofDisproof threatmate
	= state.testThreatmateState(root_conf.threatmate_node);
      if (threatmate.isCheckmateSuccess())
      {
	record->moves.push_back(*p);
	pool.add(*p, raw_king);
      }
      else if (threatmate.isUnknown())
      {
	record->result = NOT_BRINKMATE;
      }

      state.pop();
    }
    if (in_check)
      std::sort(record->moves.begin(), record->moves.end(), 
		move_order::CaptureEstimation(state.state()));
    if (record->moves.empty())
    {
      if (in_check)
	record->result = INVERSE_CHECKMATE;
      else
	record->result = NOT_BRINKMATE;
    }
  }
  record->conf.checkmate_node = std::max(record->conf.checkmate_node,
					 root_conf.checkmate_node);
  record->conf.defense_checkmove_count
    = std::max(record->conf.defense_checkmove_count,
	       root_conf.defense_checkmove_count);
  record->conf.defense_checkmate_node
    = std::max(record->conf.defense_checkmate_node,
	       root_conf.defense_checkmate_node);
}

bool osl::brinkmate::
Searcher::attack(BrinkmateRecord *record, Move todo)
{
  if (verbose >= report_do_undo)
    std::cerr << std::string(state.depth(),'+') << todo << "\n";
  const Move last_move = state.lastMove();
  
  record->is_visited = true;
  const bool take_back 
    = state.lastMove().isNormal() && (last_move.to() == todo.to()) 
    && (! state.state().hasEffectBy(alt(todo.player()), todo.to()));
  const bool in_check = EffectUtil::isKingInCheck(state.turn(), state.state());
  state.push(todo, take_back || in_check);
  const BrinkmateResult result = defense();
  state.pop();
  record->is_visited = false;

  assert(isSearchResult(result));
  record->result = betterForAttack(record->result, result);

  if (isAttackSuccess(result))
  {
    if (verbose >= report_do_undo)
      std::cerr << "threat\n";
    record->best_move = todo;
    if (last_move.isNormal())
      killer_moves.setMove(last_move, todo);
    return true;
  }
  return false;
}

osl::brinkmate::BrinkmateResult osl::brinkmate::
Searcher::attack()
{
  ++count;
  assert(state.turn() == table.attacker);
  if (stop_flag && (*stop_flag == Alarm::TIMEOUT))
    throw TableFull();
  
  if (state.isWinningState(root_conf.checkmate_node))
  {
    BrinkmateRecord *record = table.allocate(state.state().getHash());
    record->result = CHECKMATE;
    record->conf.depth = max_threshold;
    record->conf.defense_checkmove_count = max_threshold;
    record->conf.defense_checkmate_node = max_threshold;
    return record->result;
  }
  const int depth_left = depthLeft();
  if (depth_left <= 0)
    return NOT_BRINKMATE;
  BrinkmateRecord *record = table.allocate(state.state().getHash());
  assert(record);
  record->visit_count++;
  if (record->is_visited)
    return LOOP_DETECTION;
  if (provedByRecord(*record))
    return record->result;
  if (disprovedByRecord(*record))
    return record->result;
  setUpAttack(record);

  if (record->best_move.isNormal() && attack(record, record->best_move))
    return record->result;
  MoveVector killers;
#ifdef BRINKMATE_BIGRAM_KILLERS
  killer_moves.getMove(state.state(), state.lastMove(), killers);
  for (MoveVector::const_iterator p=killers.begin(); p!=killers.end(); ++p)
  {
    if ((*p != record->best_move)
	&& (std::find(record->moves.begin(), record->moves.end(), *p)
	    != record->moves.end()))
    {
      if (attack(record, *p))
	return record->result;
    }
  }
#endif
  for (BrinkmateRecord::vector_t::const_iterator p=record->moves.begin(); 
       p!=record->moves.end(); ++p)
  {
    if (record->best_move == *p)
      continue;
    if (killers.isMember(*p))
      continue;
    if (attack(record, *p))
      return record->result;
  }
  
  return record->result;
}

void osl::brinkmate::
Searcher::setUpDefense(BrinkmateRecord *record)
{
  const int depth_left = depthLeft();
  if ((record->conf.threatmate_node < root_conf.threatmate_node)
      || (record->conf.checkmate_node < root_conf.checkmate_node)
      || (record->conf.defense_checkmove_count 
	  < root_conf.defense_checkmove_count)
      || (record->conf.defense_checkmate_node 
	  < root_conf.defense_checkmate_node))
  {
    record->conf.depth = depth_left; // may need to decrease
    if (! record->moves.empty())
      record->result = WORST_DEFENSE;
  }
  else
  {
    record->conf.depth = std::max(record->conf.depth, depth_left);
  }
  if (record->moves.empty())
  {
    MoveVector all_moves;
    LegalMoves::generate(state.state(), all_moves);
    for (MoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end(); ++p)
    {
      if (state.isEffectiveDefense(*p))
      {
	record->moves.push_back(*p);
      }
    }
    std::sort(record->moves.begin(), record->moves.end(), 
	      move_order::CaptureEstimation(state.state()));
  }
  record->conf.threatmate_node = std::max(record->conf.threatmate_node,
					 root_conf.threatmate_node);
  record->conf.checkmate_node = std::max(record->conf.checkmate_node,
					 root_conf.checkmate_node);
  record->conf.defense_checkmove_count
    = std::max(record->conf.defense_checkmove_count,
	       root_conf.defense_checkmove_count);
  record->conf.defense_checkmate_node
    = std::max(record->conf.defense_checkmate_node,
	       root_conf.defense_checkmate_node);

  record->result = record->moves.empty() ? CHECKMATE : WORST_DEFENSE;
}

bool osl::brinkmate::
Searcher::defense(BrinkmateRecord *record, Move todo)
{
  using namespace move_classifier;
  const bool is_check
    = PlayerMoveAdaptor<Check>::isMember(state.state(), todo);
  const Move last_move = state.lastMove();
  const Piece last_piece = state.state().getPieceOnBoard(last_move.to());
  const Player defender = todo.player();
  bool is_sacrifice = false;
  {
    const int d = state.state().countEffect(defender, todo.to())
      + (todo.isDrop() ? 1 : 0);
    const int a = state.state().countEffect(alt(defender), todo.to());
    if (d <= 1)
    {
      if (a - d >= 0)
	is_sacrifice = true;
    }
    else if ((d == 2) && todo.isDrop())
    {
      const Position king = state.state().getKingPosition(defender);
      if (Neighboring8::isNeighboring8(king, todo.to())
	  && (a + AdditionalEffect::count2(state.state(), todo.to(), 
					   alt(defender)) >= 2))
	is_sacrifice = true;
    }
  }

  const bool is_check_or_similar = is_check || is_sacrifice;
  if (is_check_or_similar && 
      (root_conf.defense_checkmove_count <= state.defenseCheckCount()))
  {
    record->result = betterForDefense(record->result, BRINKMATE);
    return false;
  }

  if (verbose >= report_do_undo)
    std::cerr << std::string(state.depth(),'+') << todo << "\n";
  record->is_visited = true;
  state.push(todo, is_check_or_similar);
  const BrinkmateResult result = attack();
  state.pop();
  record->is_visited = false;

  assert(isSearchResult(result));
  record->result = betterForDefense(record->result, result);

  if (! isAttackSuccess(result))
  {
    if (verbose >= report_do_undo)
      std::cerr << "defend\n";
    if (last_move.isNormal())
      killer_moves.setMove(last_move, todo);
    record->best_move = todo;
    return true;
  }
  return false;
}


osl::brinkmate::BrinkmateResult osl::brinkmate::
Searcher::defense()
{
  ++count;
  assert(alt(state.turn()) == table.attacker);
  if (state.isWinningState(0))
    return INVERSE_CHECKMATE;
  const int depth_left = depthLeft();
  const bool in_check = EffectUtil::isKingInCheck(state.turn(), state.state());

  if (depth_left <= 0)
  {
    if (in_check)
    {
      MoveVector all_moves;
      LegalMoves::generate(state.state(), all_moves);
      if (all_moves.empty())
	return CHECKMATE;	// XXX: stalemate
    }
    return NOT_BRINKMATE;
  }
  
  BrinkmateRecord *record = table.allocate(state.state().getHash());
  assert(record);
  record->visit_count++;
  if (record->is_visited)
    return LOOP_DETECTION;
  if (provedByRecord(*record))
    return record->result;
  if (disprovedByRecord(*record))
    return record->result;

  setUpDefense(record);

  if (record->best_move.isNormal() && defense(record, record->best_move))
    return record->result;
  MoveVector killers;
#ifdef BRINKMATE_BIGRAM_KILLERS
  killer_moves.getMove(state.state(), state.lastMove(), killers);
  for (MoveVector::const_iterator p=killers.begin(); p!=killers.end(); ++p)
  {
    if ((*p != record->best_move)
	&& (std::find(record->moves.begin(), record->moves.end(), *p)
	    != record->moves.end()))
    {
      if (defense(record, *p))
	return record->result;
    }
  }
#endif
  for (BrinkmateRecord::vector_t::const_iterator p=record->moves.begin(); 
       p!=record->moves.end(); ++p)
  {
    if (record->best_move == *p)
      continue;
    if (killers.isMember(*p))
      continue;
    if (defense(record, *p))
      return record->result;
  }

  const ProofDisproof attack_back
    = state.testWinningState(root_conf.defense_checkmate_node);
  if (attack_back.isCheckmateSuccess())
  {
    if (verbose >= report_do_undo)
      std::cerr << "attack_back\n";
    record->result = INVERSE_CHECKMATE;
    return record->result;
  }
  else if (attack_back.isUnknown())
  {
    record->result = betterForDefense(record->result, BRINKMATE);
  }
  if ((record->result == CHECKMATE) && (! in_check))
    record->result = BRINKMATE;
  return record->result;
}

bool osl::brinkmate::
Searcher::provedByRecord(const BrinkmateRecord& record) const
{
  if (record.result == CHECKMATE)
    return true;
  if (record.result != BRINKMATE)
    return false;
  return (record.conf.depth >= depthLeft())
    && (record.conf.defense_checkmove_count 
	>= root_conf.defense_checkmove_count)
    && (record.conf.defense_checkmate_node
	>= root_conf.defense_checkmate_node);
}

bool osl::brinkmate::
Searcher::disprovedByRecord(const BrinkmateRecord& record) const
{
  if (record.result == INVERSE_CHECKMATE)
    return true;
  if (record.result == LOOP_DETECTION)
    return true;
  if (record.result != NOT_BRINKMATE)
    return false;
  return (record.conf.depth >= depthLeft())
    && (record.conf.threatmate_node >= root_conf.threatmate_node)
    && (record.conf.checkmate_node >= root_conf.checkmate_node);
}

void osl::brinkmate::
Searcher::showBestMoves() const
{
  HashKey key = state.state().getHash();
  const BrinkmateRecord *record = table.find(key);
  while (record && (! record->moves.empty()))
  {
    Move best_move = record->best_move;
    if (! best_move.isNormal())
    {
      int max_depth = 0;
      for (BrinkmateRecord::vector_t::const_iterator p=record->moves.begin(); 
	   p!=record->moves.end(); ++p)
      {
	const HashKey new_key = key.newHashWithMove(*p);
	const BrinkmateRecord *new_record = table.find(new_key);
	if (new_record && new_record->conf.depth > max_depth)
	{
	  max_depth = new_record->conf.depth;
	  best_move = *p;
	}
      }
    }
    std::cerr << record::csa::show(best_move) 
	      << " " << record->result << " ";
    if ((record->result == CHECKMATE)
	|| (record->result == INVERSE_CHECKMATE))
      break;
    if (! best_move.isNormal())
      break;
    key = key.newHashWithMove(best_move);
    record = table.find(key);
  }
  std::cerr << "\n";
}

void osl::brinkmate::
Searcher::showTree(const HashKey& key, int depth) const
{
  const BrinkmateRecord *record = table.find(key);
  if (! record)
    return;
  if (depth >= max_depth + root_conf.defense_checkmove_count)
    return;
  std::cerr << record->result << " t " << record->conf.threatmate_node 
	    << " v " << record->visit_count 
	    << " d " << record->conf.depth 
	    << " b " << record::csa::show(record->best_move)
	    << "\n";
  if (record->best_move.isNormal())
  {
    std::cerr << std::string(depth,'*') 
	      << record::csa::show(record->best_move) 
	      << "\n";
    const HashKey new_key = key.newHashWithMove(record->best_move);
    showTree(new_key, depth+1);
  }
  else
  {
    for (BrinkmateRecord::vector_t::const_iterator p=record->moves.begin(); 
	 p!=record->moves.end(); ++p)
    {
      std::cerr << std::string(depth,'*') 
		<< record::csa::show(*p) 
		<< "\n";
      const HashKey new_key = key.newHashWithMove(*p);
      showTree(new_key, depth+1);
    }
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
