/* effectUtil.cc
 */
#include "osl/effect_util/effectUtil.h"
#include "osl/effect_util/effectUtil.tcc"
#include "osl/eval/pieceEval.h"
#include "osl/state/numEffectState.h"

namespace osl
{
  template Piece 
  EffectUtil::cheapestPiece(Player, const NumEffectState&, Position);
  template Piece 
  EffectUtil::promotablePiece(Player, const NumEffectState&, Position);
  template void
  EffectUtil::showEffect(const NumEffectState&, Position, std::ostream&);
  template Piece
  EffectUtil::safeCaptureNotByKing(Player, const NumEffectState&, Position);
  template void
  EffectUtil::findEffect(Player, const NumEffectState&, Position, PieceVector&);
  template bool
  EffectUtil::isOpenAttackMove<BLACK>(const SimpleState&, Position, Position, Position);
  template bool
  EffectUtil::isOpenAttackMove<WHITE>(const SimpleState&, Position, Position, Position);
  template bool
  EffectUtil::isAddEffectMove<BLACK>(const SimpleState&, Position, Move);
  template bool
  EffectUtil::isAddEffectMove<WHITE>(const SimpleState&, Position, Move);
#ifdef NEED_EXTRA_TEMPLATE_INSTANTIATION
  template Piece 
  EffectUtil::safeCaptureNotByKing<BLACK, NumEffectState>(NumEffectState const&, Position, Piece);
  template Piece 
  EffectUtil::safeCaptureNotByKing<WHITE, NumEffectState>(NumEffectState const&, Position, Piece);
#endif
}

struct osl::effect_util::EffectUtil::SelectMostValuable
{
  const SimpleState& state;
  Player target;
  Position move_from;
  Piece result;
  int result_value;
  SelectMostValuable(const SimpleState& s, Player t, Position f) 
    : state(s), target(t), move_from(f),
      result(Piece::EMPTY()), result_value(0)
  {
  }
  void operator()(Position pos)
  {
    if (pos == move_from)
      return;
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() == target)
    {
      const Ptype ptype = cur.ptype();
      const int curValue 
	= eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE,ptype));
      assert(curValue > 0);
      if ((ptype != KING) && (curValue > result_value))
      {
	result = cur;
	result_value = curValue;
      }
    }
  }
};

osl::Piece osl::EffectUtil::
mostValuableEffectOfMoveNotKing(Player target, const SimpleState& state, Move m)
{
  SelectMostValuable selector(state, target, m.from());
  forEachEffectPieceOfMove(state, m, selector);
  return selector.result;
}

template <int N>
struct osl::effect_util::EffectUtil::SelectMostValuableN
{
  const SimpleState& state;
  Player target;
  Position move_from;
  CArray<Piece,N>& result;
  CArray<int,N> result_value;
  SelectMostValuableN(const SimpleState& s, Player t, Position f,
		      CArray<Piece,N>& out)
    : state(s), target(t), move_from(f), result(out)
  {
    result.fill(Piece::EMPTY());
    result_value.fill(0);
  }
  void operator()(Position pos)
  {
    if (pos == move_from)
      return;
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() == target)
    {
      const Ptype ptype = cur.ptype();
      const int curValue
	= eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE,ptype));
      assert(curValue > 0);
      if (ptype == KING)
	return;
      for (int i=0; i<N; ++i)
      {
	if (curValue > result_value[i])
	{
	  for (int j=i; j+1<N; ++j)
	  {
	    result[j+1] = result[j];
	    result_value[j+1] = result_value[j];
	  }
	  result[i] = cur;
	  result_value[i] = curValue;
	  return;
	}
      }
    }
  }
};

void osl::EffectUtil::
mostValuableEffectOfMoveNotKing(Player target, const SimpleState& state, 
				Move m, CArray<Piece,2>& out)
{
  SelectMostValuableN<2> selector(state, target, m.from(), out);
  forEachEffectPieceOfMove(state, m, selector);
}

struct osl::effect_util::EffectUtil::FindThreat
{
  const NumEffectState& state;
  Player target;
  int attacker_value;
  PieceVector& supported, & unsupported;
  FindThreat(const NumEffectState& st, Player t, int a,
	     PieceVector& s, PieceVector& u)
    : state(st), target(t), attacker_value(a), supported(s), unsupported(u)
  {
  }
  void operator()(Position pos)
  {
    const Piece cur = state.getPieceOnBoard(pos);
    assert(cur.isPiece());
    if (cur.owner() != target)
      return;
    if (state.hasEffectBy(target, pos))
    {
      if (abs(eval::PieceEval::captureValue(cur.ptypeO()))
	  > attacker_value)
	supported.push_back(cur);
    }
    else
    {
      unsupported.push_back(cur);
    }
  }
};

void osl::EffectUtil::
findThreat(const NumEffectState& state, Position position,
	   PtypeO ptypeo, PieceVector& out)
{
  PieceVector supported, unsupported;
  const int attacker_value = abs(eval::PieceEval::captureValue(ptypeo));
  FindThreat f(state, alt(getOwner(ptypeo)), attacker_value, 
	       supported, unsupported);
  forEachEffectOfPtypeO<NumEffectState, FindThreat, false>
    (state, position, ptypeo, f);

  unsupported.sortByPtype();
  supported.sortByPtype();
  PieceVector::iterator u=unsupported.begin(), s=supported.begin();

  if (u!=unsupported.end())
  {
    while ((s!=supported.end()) 
	   && ((abs(eval::PieceEval::captureValue(s->ptypeO()))
		- attacker_value)
	       > abs(eval::PieceEval::captureValue(u->ptypeO()))))
    {
      out.push_back(*s);
      ++s;
    }
  }
  out.push_back(u, unsupported.end());
  out.push_back(s, supported.end());
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
