/* searchPlayer.cc
 */
#include "osl/game_playing/searchPlayer.h"
#include "osl/game_playing/searchPlayer.tcc"
#include "osl/game_playing/historyToTable.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/tableStack.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/timeControl.h"
#include "osl/search/searchEvalBase.h"
#include "osl/search/bigramKillerMove.h"
#include "osl/checkmate/checkmateSearcher.h"
#include "osl/enter_king/enterKing.h"
#include "osl/progress/effect5x3.h"
#include "osl/container/moveStack.h"
#include "osl/misc/realTime.h"
#include <cmath>
#include <ctime>
#include <iostream>

osl::game_playing::SearchPlayer::
Config::Config()
  : limit(1200), node_limit(200000), table_size(10000),
    table_record_limit(-5), initial_limit(600), deepening_step(200),
    total_checkmate_limit(CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT),
    verbose(false)
{
}

bool osl::game_playing::operator==(const SearchPlayer::Config& l, 
				   const SearchPlayer::Config& r)
{
  return (l.limit == r.limit) && (l.node_limit == r.node_limit)
    && (l.table_size == r.table_size) 
    && (l.table_record_limit == r.table_record_limit)
    && (l.initial_limit == r.initial_limit)
    && (l.deepening_step == r.deepening_step)
    && (l.total_checkmate_limit == r.total_checkmate_limit)
    && (l.verbose == r.verbose);
}

osl::game_playing::
SearchPlayer::SearchPlayer()
  : recorder_ptr(new SearchRecorder("/dev/null")),
    bigram_ptr(new search::BigramKillerMove()),
    table_stack_ptr(new TableStack()),
    search_result_ptr(new SearchResult()),
    stop_flag(0),
    copy_history_table(false)
{
}

osl::game_playing::
SearchPlayer::SearchPlayer(const SearchPlayer& copy)
  : config(copy.config),
    recorder_ptr(new SearchRecorder("/dev/null")),
    bigram_ptr(new search::BigramKillerMove()),
    table_stack_ptr(),
    search_result_ptr(new SearchResult()),
    stop_flag(0)
{
  if (copy.table_stack_ptr)
    table_stack_ptr.reset(new TableStack(*copy.table_stack_ptr));
}

osl::game_playing::
SearchPlayer::~SearchPlayer()
{
}

void osl::game_playing::
SearchPlayer::enableTableStack(bool enable)
{
  if (enable)
  {
    if (! table_stack_ptr)
      table_stack_ptr.reset(new TableStack());
  }
  else
  {
    table_stack_ptr.reset();
  }
}

void osl::game_playing::
SearchPlayer::swapTable(SearchPlayer& other)
{
  table_ptr.swap(other.table_ptr);
}

bool osl::game_playing::
SearchPlayer::canStopSearch()
{
  return stop_flag.notify;
}

bool osl::game_playing::
SearchPlayer::stopSearchNow()
{
  if (! stop_flag.notify)
  {
    std::cerr << "SearchPlayer stop_flag is null\n";
    return false;
  }
  *stop_flag.notify = Alarm::TIMEOUT;
  return true;
}

void osl::game_playing::
SearchPlayer::setTotalCheckmateLimit(int limit)
{
  config.total_checkmate_limit = limit;
}

void osl::game_playing::
SearchPlayer::resetRecorder(search::SearchRecorder *new_recorder)
{
  recorder_ptr.reset(new_recorder);
}

int osl::game_playing::
SearchPlayer::setUpTable(const GameState& gs, int draw)
{
  misc::RealTime timer(0);
  time_t t = time(0);
  if (table_ptr.get() && table_ptr->isVerbose())
    std::cerr << "setUpTable " << ctime(&t) << std::flush;
  
  const int black_win = search::FixedEval::winByLoop(BLACK);
  const int white_win = search::FixedEval::winByLoop(WHITE);
  if (! table_ptr.get())
    table_ptr.reset(new SimpleHashTable(config.table_size, 
					config.table_record_limit, config.verbose));
  else
    table_ptr->clear();
  table_ptr->setVerbose(config.verbose);
  if (table_stack_ptr)
    table_stack_ptr->migrate(*table_ptr);

  HistoryToTable::adjustTable(gs, *table_ptr, black_win, draw, white_win);

  checkmate_ptr.reset(new CheckmateSearcher(config.total_checkmate_limit));
  checkmate_ptr->writeRootHistory(gs.counter(), gs.moveHistory(),
				  gs.state(), gs.state().getTurn());

  if (table_ptr->isVerbose())
  {
    t = time(0);
    std::cerr << "setup done " << ctime(&t) << std::flush;
  }
  return static_cast<int>(ceil(timer.getConsumedInDouble()));
}

void osl::game_playing::
SearchPlayer::setDepthLimit(int l, int il, int ds)
{
  config.limit = l;
  config.initial_limit = il;
  config.deepening_step = ds;
}

void osl::game_playing::
SearchPlayer::setNodeLimit(int nl)
{
  config.node_limit = nl;
}

void osl::game_playing::
SearchPlayer::setTableLimit(int size, int record_limit)
{
  config.table_size = size;
  config.table_record_limit = record_limit;

  table_ptr.reset();
}

void osl::game_playing::
SearchPlayer::setVerbose(bool v)
{
  config.verbose = v;
}

void osl::game_playing::
SearchPlayer::pushMove(Move move)
{
  if (table_stack_ptr)
  {
    if (table_ptr.get() && (move == search_result_ptr->best_move))
    {
      table_stack_ptr->pushMove(move, search_result_ptr->root, *table_ptr);
    }
    else
    {
      table_stack_ptr->pushMove(move);
    }
  }
}
void osl::game_playing::
SearchPlayer::popMove()
{
  if (table_stack_ptr)
    table_stack_ptr->popMove();
}

int osl::game_playing::
SearchPlayer::secondsForThisMove(const GameState& state, int total_seconds_left,
				 int byoyomi)
{
  if (byoyomi < 0)
    return -1; // ̵

  if (total_seconds_left < byoyomi)
    return byoyomi;

  if (byoyomi == 0)
  {
    // 330ؤ褦ˤ
    // http://www32.ocn.ne.jp/~yss/csa14rep.html
    // 90ä1þΤ240ɤ (330-90)
    total_seconds_left -= (240 - static_cast<int>(state.moveHistory().size()));
  }
  const int seconds_for_this_move 
    = ((total_seconds_left <= 1) 
       ? 1
       : search::TimeControl::secondsForThisMove(total_seconds_left));

  // Think more if book leads us to near endgame
  if (total_seconds_left >= 600)
  {
    const progress::Effect5x3 progress(state.state());
    if ((progress.progress16().value() >= 15)
	&& ((progress.progress16(BLACK).value() >= 13)
	    || (progress.progress16(WHITE).value() >= 13)))
      return seconds_for_this_move*2;
  }
  return std::max(byoyomi, seconds_for_this_move);
}

const osl::Move osl::game_playing::
SearchPlayer::selectBestMove(const GameState& state, int total_seconds_left,
			     int byoyomi)
{
  if (EnterKing::canDeclareWin(state.state()))
    return Move::DeclareWin();
  const int seconds_for_this_move
    = secondsForThisMove(state, total_seconds_left, byoyomi);
  return searchWithSecondsForThisMove(state, seconds_for_this_move);
}

void osl::game_playing::
SearchPlayer::pickUpBestMoves(const GameState& game_state, MoveVector& out) const
{
  if (!table_ptr)
    return;

  const HashKey& key = game_state.state().getHash();
  const SimpleHashRecord *record = table_ptr->find(key);
  if (! record)
    return;

  if (record->qrecord.bestMove().isNormal())
    out.push_back(record->qrecord.bestMove());

  const Player turn = game_state.state().getTurn();

  int best_value = search::FixedEval::winByCheckmate(alt(turn));
  Move best_move = Move::INVALID();
  for (int limit=600; limit<=1200; limit+=200)
  {
    for (SearchMoveSet::const_iterator it=record->moves().begin();
	 it != record->moves().end(); it++)
    {
      if (! it->record)
	continue;
      if (out.isMember(it->getMove()))
	continue;
      if (! it->getMove().isNormal())
	continue;
      if (it->getLogProb() > limit)
	continue;
      if (it->record->hasLowerBound(0)
	  && eval::betterThan(turn, it->record->lowerBound(), best_value) 
	  && (! out.isMember(it->getMove())))
      {
	best_move = it->getMove();
	best_value = it->record->upperBound();
      }
    }
    if (best_move.isNormal())
    {
      out.push_back(best_move);
      break;
    }
  }
  if (record->bestMove().getMove().isNormal())
    out.push_back(record->bestMove().getMove());
  out.unique();
}

#ifdef USE_NTESUKI
osl::game_playing::SearchPlayer::
NtesukiThread::NtesukiThread(Move& next_move,
	      volatile bool *thread_finished,
	      volatile bool *stop_flag,
	      NumEffectState state)
  : next_move(next_move), thread_finished(thread_finished),
    stop_flag(stop_flag), state(state)
{
}

void osl::game_playing::SearchPlayer::
NtesukiThread::operator()()
{
  std::cerr << "start ntesuki search\n";
  *thread_finished = false;

  const Player P = state.getTurn();
  const HashKey key = osl::HashKey::calcHash(state);;

  boost::scoped_ptr<osl::ntesuki::NtesukiAttackMoveGenerator>
    gam(new osl::ntesuki::GetAttackMoves());
  boost::scoped_ptr<osl::ntesuki::NtesukiDefenseMoveGenerator>
    gdm(new osl::ntesuki::GetDefenseMoves());

  osl::ntesuki::NtesukiSearcher
    searcher(state, gam.get(), gdm.get(), 500000u, stop_flag, true, 2);
   
  try
  {
    int ntesuki_num = searcher.searchSlow(P, 10000000);
    if (-1 != ntesuki_num)
    {
      const osl::ntesuki::PdNtesukiTable& table
	= searcher.getTableSlow(P);
      const osl::ntesuki::PdNtesukiRecord *record
	= table.find(key);
      next_move = record->getBestMove(ntesuki_num).getMove();
    }
  }
  catch (ntesuki::ReadNodeLimit& e)
  {
  }
  catch (ntesuki::TableFull& e)
  {
  }
  catch (std::runtime_error& e)
  {
    std::cerr << e.what() << "\n";
  }
  std::cerr << "end ntesuki search\n";
  *thread_finished = true;
}
#endif
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
