/* speculativeSearchPlayer.cc
 */
#include "osl/game_playing/speculativeSearchPlayer.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/searchPlayer.h"
#include "osl/search/timeControl.h"
#include "osl/search/searchRecorder.h"
#include "osl/container/moveStack.h"
#include "osl/record/csa.h"
#include "osl/sennichite.h"
#include <boost/thread.hpp>
#include <iostream>
#include <ctime>
#include <unistd.h>

#define DEBUG_SPECULATIVE_EXECUTION
#define EXTRA_DEBUG_SPECULATIVE_EXECUTION

struct osl::game_playing::SpeculativeSearchPlayer::SpeculativeThread
{
  boost::shared_ptr<boost::thread> thread;
  boost::shared_ptr<GameState> state;
  boost::shared_ptr<SearchPlayer> player;
  bool parent_is_verbose;
  volatile bool thread_finished;
  Move predicted_move;
  Move next_move;
  bool verbose;
  // operator()񤭹
  Move *next_move_for_thread;
  volatile bool *thread_finished_for_thread;

  SpeculativeThread() : player(), parent_is_verbose(false),
			thread_finished(true),
			verbose(false),
			next_move_for_thread(&next_move),
			thread_finished_for_thread(&thread_finished)
  {
  }
  void operator()()
  {
    // boost::thread ˤϤobject *copy* 
    // ϤΤǡǤϥСѿɤ߽ФǤʤ
    // i.e., ͤѹƤФˡʤ
    // äƽ񤯻ˤϥݥ󥿷ͳǡ
    assert(predicted_move.isNormal());
    *thread_finished_for_thread = false;
    state->pushMove(predicted_move);
    player->pushMove(predicted_move);
    player->setVerbose(verbose);
    *next_move_for_thread = Move::INVALID();
    *next_move_for_thread = player->selectBestMove(*state, -1, -1); // infinity
    *thread_finished_for_thread = true;
#ifdef DEBUG_SPECULATIVE_EXECUTION
    time_t now = time(0);
    std::cerr << "if " << record::csa::show(predicted_move) 
	      << " then " << record::csa::show(*next_move_for_thread) << " " << ctime(&now);
#endif
    if (parent_is_verbose)
      player->recorder().reportCount(std::cerr);
  }
  void run(Move prediction, const GameState& main_state,
	   const SearchPlayer& main_player, bool verbose)
  {
    player.reset(dynamic_cast<SearchPlayer*>(main_player.clone()));
    parent_is_verbose = player->getConfig().verbose;
    assert(! thread.get());
    predicted_move = prediction;
    state = main_state.clone();
    this->verbose = verbose;
    try
    {
      thread.reset(new boost::thread(*this));
#ifdef DEBUG_SPECULATIVE_EXECUTION
      time_t now = time(0);
      std::cerr << "speculative search for " << record::csa::show(predicted_move) 
		<< " " << ctime(&now);
#endif
    }
    catch (boost::thread_resource_error&)
    {
      time_t now = time(0);
      std::cerr << "thread creation failure " << ctime(&now);
      thread.reset();
      predicted_move = Move::INVALID();
    }
  }
  void join()
  {
    if (thread.get())
    {
      thread->join();
      thread.reset();
    }
    predicted_move = Move::INVALID();
  }
  void stopSearchNow()
  {
    if (! thread_finished)
    {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
      time_t now = time(0);
      std::cerr << "need to call stopSearchNow " << ctime(&now);
#endif
      while ((! thread_finished)
	     && (! player->canStopSearch()))	// TODO:conditional wait
      {
	timespec rq = { 0, 20000000 }, rm;
	while (nanosleep(&rq, &rm) == -1) 
	{
	  rq = rm;
	}
      }
      if (! thread_finished)
      {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
	time_t now = time(0);
	std::cerr << "calling stopSearchNow " << ctime(&now);
#endif
	const bool success
	  = player->stopSearchNow();
	if (! success)
	{
	  std::cerr << "stop search failed" << this << "\n";
	}
      }
    }
    join();
  }
};

osl::game_playing::SpeculativeSearchPlayer::
SpeculativeSearchPlayer(Player my_turn, SearchPlayer *player, int threads)
  : main_player(player), 
    speculative_thread0(new SpeculativeThread()),
    speculative_thread1(new SpeculativeThread()),
    my_turn(my_turn),
    max_threads(threads)
{
}
osl::game_playing::
SpeculativeSearchPlayer::~SpeculativeSearchPlayer()
{
}

osl::game_playing::ComputerPlayer* 
osl::game_playing::SpeculativeSearchPlayer::clone() const
{
  return new SpeculativeSearchPlayer(my_turn, 
				     dynamic_cast<SearchPlayer*>(main_player->clone()));
}

void osl::game_playing::SpeculativeSearchPlayer::
pushMove(Move m)
{
  main_player->pushMove(m);
  if (m.player() == my_turn)
  {
    if (previous_state.get() && speculative_search_allowed)
    {
      assert(speculative_thread0.get());
      assert(speculative_thread1.get());
      previous_state->pushMove(m);
      MoveVector moves;
      main_player->pickUpBestMoves(*previous_state, moves);
      const bool run_thread0 = moves.size() > 0 && max_threads > 0;
      const bool run_thread1 = moves.size() > 1 && max_threads > 1;
      if (run_thread0)
	speculative_thread0->run(moves[0], *previous_state, *main_player,
				 (! run_thread1));
      if (run_thread1)
	speculative_thread1->run(moves[1], *previous_state, *main_player,
				 false);
      previous_state.reset();
    }
  }
  else
  {
    if (speculative_thread0->predicted_move != m)
    {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
      if (speculative_search_allowed)
      {
	time_t now = time(0);
	std::cerr << "stop thread0 " << ctime(&now);
      }
#endif
      speculative_thread0->stopSearchNow();
    }
    if (speculative_thread1->predicted_move != m)
    {
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
      if (speculative_search_allowed)
      {
	time_t now = time(0);
	std::cerr << "stop thread1 " << ctime(&now);
      }
#endif
      speculative_thread1->stopSearchNow();
    }
  }
}

void osl::game_playing::SpeculativeSearchPlayer::
popMove()
{
  main_player->popMove();
  previous_state.reset();
  if (! speculative_thread0->thread_finished)
    speculative_thread0->stopSearchNow();
  if (! speculative_thread1->thread_finished)
    speculative_thread1->stopSearchNow();
}

bool osl::game_playing::SpeculativeSearchPlayer::
stopSearchNow()
{
  return main_player->stopSearchNow();
}

const osl::Move osl::game_playing::SpeculativeSearchPlayer::
selectBestMove(const GameState& state, int seconds, int byoyomi)
{
  Move result = Move::INVALID();
  SpeculativeThread *match = 0;
  const Move last_move = state.moveHistory().lastMove();
  if (last_move.isNormal())
  {
    if (speculative_thread0->predicted_move == last_move)
      match = speculative_thread0.get();
    else if (speculative_thread1->predicted_move == last_move)
      match = speculative_thread1.get();
  }
  
  if (match)
  {
    time_t start_time = time(0);
    time_t stop_time = start_time
      + std::max(1,
		 match->player->secondsForThisMove(state, seconds, byoyomi)-1);
#ifdef DEBUG_SPECULATIVE_EXECUTION
    std::cerr << "match " << ctime(&start_time);
    std::cerr << "wait for " << ctime(&stop_time);
#endif
    while (true)
    {
      if (match->thread_finished)
	break;

      time_t now = time(0);
      if (now >= stop_time)
      {
	match->stopSearchNow();
#ifdef EXTRA_DEBUG_SPECULATIVE_EXECUTION
	std::cerr << "stop search (match): " << ctime(&now);
#endif
	break;
      }
      timespec rq = { 1, 0 }, rm;
      nanosleep(&rq, &rm);
    }
    result = match->next_move;
    main_player->swapTable(*match->player);
#ifdef DEBUG_SPECULATIVE_EXECUTION
    time_t now = time(0);
    std::cerr << "returned " << record::csa::show(result) 
	      << " " << ctime(&now);
#endif
  }
  else
  {
#ifdef DEBUG_SPECULATIVE_EXECUTION
    time_t now = time(0);
    std::cerr << "start search for " << record::csa::show(last_move)
	      << " (unexpected): " << ctime(&now);
#endif
    result = main_player->selectBestMove(state, seconds, byoyomi);
  }
  previous_state = state.clone();
  speculative_thread0->join();
  speculative_thread1->join();
  return result;
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
