/* repetitionCounter.cc
 */
#include "osl/repetitionCounter.h"
#include "osl/hash/hashKey.h"
#include "osl/state/simpleState.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/move_classifier/check.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/stl/hash_map.h"
#include <iostream>

#if (__GNUC__ < 3 || (__GNUC__ ==3 && __GNUC_MINOR__ < 4) ) && defined(USE_GPL_POOL_ALLOCATOR)
template class osl::__gnu_cxx::__pool_alloc<true,0>;
#endif

typedef osl::RepetitionCounter::list_t list_t;
typedef osl::hash_map<osl::HashKey,list_t> map_t;

struct osl::RepetitionCounter::Table : public map_t
{
};

void osl::RepetitionCounter::
clear() 
{
  table->clear();
  continuous_check[0].clear();
  continuous_check[1].clear();
  hash_history.clear();
  
  continuous_check[0].push_back(0);
  continuous_check[1].push_back(0);
}

osl::RepetitionCounter::
RepetitionCounter() : table(new Table()), hash_history(255)
{
  clear();
}

osl::RepetitionCounter::
RepetitionCounter(const RepetitionCounter& c)
  : continuous_check(c.continuous_check), 
    hash_history(c.hash_history)
{
  if (c.table)
    table.reset(new Table(*c.table));
  assert(isConsistent());
}


osl::RepetitionCounter::
RepetitionCounter(const HashEffectState& initial) 
  : table(new Table()), hash_history(255)
{
  clear();
  push(initial);
}

osl::RepetitionCounter::
RepetitionCounter(const NumEffectState& initial)
  : table(new Table()), hash_history(255)
{
  clear();
  const HashKey key = HashKey::calcHash(initial);
  push(key, initial);
}

osl::RepetitionCounter::
~RepetitionCounter()
{
  assert(isConsistent());
}

void osl::RepetitionCounter::
push(const HashKey& new_key, bool is_check)
{
  const Player last_turn = alt(new_key.turn());
  if (is_check)
  {
    continuous_check[playerToIndex(last_turn)].push_back(checkCount(last_turn)+1);
  }
  else
  {
    continuous_check[playerToIndex(last_turn)].push_back(0);
  }

  const Table::iterator p=table->find(new_key);
  if (p == table->end())
  {
    (*table)[new_key].push_front(order());
  }
  else
  {
    list_t& l = p->second;
    l.push_front(order());
  }
  hash_history.push(new_key);
}

void osl::RepetitionCounter::
push(const HashKey& key, const NumEffectState& state)
{
  const bool is_check
    = EffectUtil::isKingInCheck(state.getTurn(), state);
  push(key, is_check);
}

void osl::RepetitionCounter::
push(const HashEffectState& state)
{
  push(state.getHash(), state);
}

void osl::RepetitionCounter::
push(const HashEffectState& state, Move move)
{
  assert(move.isValidOrPass());
  assert(state.getTurn() == move.player());
  
  HashKey key = state.getHash();
  key = key.newHashWithMove(move);

  // ؤβ򸡽
  using namespace move_classifier;
  const bool is_check 
    = (!move.isPass()) && PlayerMoveAdaptor<Check>::isMember(state, move);
  push(key, is_check);
}

void osl::RepetitionCounter::
pop()
{
  assert(order());
  const HashKey last_key = hash_history.top();
  hash_history.pop();
  
  const Player last_turn = alt(last_key.turn());
  assert(! continuous_check[playerToIndex(last_turn)].empty());
  continuous_check[playerToIndex(last_turn)].pop_back();

  const Table::iterator p=table->find(last_key);
  assert(p != table->end());

#ifndef NDEBUG
  const list_t::iterator q = p->second.begin();
  assert(q != p->second.end());
  assert(*q == order());
#endif
  p->second.pop_front();
  if (p->second.empty())
    table->erase(p);
}

int osl::RepetitionCounter::
getLastMove(const HashKey& key) const
{
  const Table::const_iterator p=table->find(key);
  if (p == table->end())
    return -1;
  return p->second.front();
}
int osl::RepetitionCounter::
getFirstMove(const HashKey& key) const
{
  const Table::const_iterator p=table->find(key);
  if (p == table->end())
    return -1;
  list_t::const_iterator q = p->second.begin();
  assert(q != p->second.end());
  int result = *q++;
  while (q != p->second.end())
    result = *q++;
  return result;
}

const osl::Sennichite osl::RepetitionCounter::
isSennichite(const HashEffectState& state, Move move) const
{
  HashKey key = state.getHash();
  key = key.newHashWithMove(move);
  const Table::const_iterator p=table->find(key);
  if (p == table->end())
    return Sennichite::NORMAL();

  // 3ȼ4
  if (p->second.size() < 3)
    return Sennichite::NORMAL();
  return isAlmostSennichite(key);
}

unsigned int osl::RepetitionCounter::
countRepetition(const HashKey& key) const
{
  const Table::const_iterator p=table->find(key);
  if (p == table->end())
    return 0;
  return p->second.size();
}

const list_t osl::RepetitionCounter::
getRepetitions(const HashKey& key) const
{
  Table::const_iterator p=table->find(key);
  if (p == table->end())
    return list_t();
  return p->second;
}

void osl::RepetitionCounter::
printMatches(const HashKey& key) const
{
  Table::const_iterator p=table->find(key);
  if (p == table->end())
    return;
  const list_t& l = p->second;
  for (list_t::const_iterator q=l.begin(); q!=l.end(); ++q)
  {
    std::cerr << *q << " ";
  }
  std::cerr << "\n";
}

bool osl::RepetitionCounter::
isConsistent() const
{
  HashKeyStack history = hash_history;
  Table table(*this->table);
  CArray<osl::vector<int>, 2> continuous_check = this->continuous_check;
  while (history.empty())
  {
    const HashKey last_key = history.top();
    history.pop();
  
    const Player last_turn = alt(last_key.turn());
    assert(! continuous_check[playerToIndex(last_turn)].empty());
    continuous_check[playerToIndex(last_turn)].pop_back();

    const Table::iterator p=table.find(last_key);
    if (p == table.end())
    {
      std::cerr << "oops, " << this << "\n";
      return false;
    }
    assert(p != table.end());

#ifndef NDEBUG
    const list_t::iterator q = p->second.begin();
    assert(q != p->second.end());
    assert(*q == order());
#endif
    p->second.pop_front();
    if (p->second.empty())
      table.erase(p);
  }
  return true;
}

bool osl::operator==(const RepetitionCounter& l, const RepetitionCounter& r)
{
  if (*l.table != *r.table)
    return false;
  if (l.continuous_check[0] != r.continuous_check[0])
    return false;
  if (l.continuous_check[1] != r.continuous_check[1])
    return false;
  return l.hash_history == r.hash_history;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
