/* SparseRegressionMultiplier.cc
 */
#include "osl/stat/sparseRegressionMultiplier.h"
#include <boost/scoped_array.hpp>
#include <algorithm>
#include <cassert>

// CREATORS
osl::stat::
SparseRegressionMultiplier::SparseRegressionMultiplier(unsigned int dimA)
  : m_dimA(dimA)
{
}

osl::stat::
SparseRegressionMultiplier::~SparseRegressionMultiplier()
{
}

// MANIPULATORS
// ACCESSORS
void osl::stat::
SparseRegressionMultiplier::newIteration() const
{
}

void osl::stat::
SparseRegressionMultiplier::prod(const double *x, double *out) const
{
  newIteration();
  std::fill(&out[0], &out[m_dimA], 0.0);

  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  boost::scoped_array<double> values(new double[m_dimA]);
  unsigned int non_zeros;
  while (true)
  {
    const bool go_next = getVectorX(non_zeros, &indices[0], &values[0]);
    addProduct(non_zeros, &indices[0], &values[0], x, out);
    if (! go_next)
      break;
  }
}

unsigned int osl::stat::SparseRegressionMultiplier::
computeXtY(DoubleReader& y, double *xty, double *diag_inv) const
{
  boost::scoped_array<unsigned int> indices(new unsigned int[dim()]);
  boost::scoped_array<double> values(new double[dim()]);
	
  std::fill(&xty[0], &xty[dim()], 0.0);
  std::fill(&diag_inv[0], &diag_inv[dim()], 0.0);

  unsigned int instance_id=0;
  unsigned int non_zeros;
  while (true)
  {
    const bool go_next = getVectorX(non_zeros, &indices[0], &values[0]);
    const double cur_y = y.read();
    for (unsigned int ip=0; ip<non_zeros; ++ip)
    {
      const unsigned int p = indices[ip];
      const double vp = values[ip];
#if 0
      if (p >= dim())
	std::cerr << p << " >= " << dim() << "\n";
#endif
      assert(p < dim());
      xty[p] += vp * cur_y;
      diag_inv[p] += vp*vp;
    }
    ++instance_id;
    if (! go_next)
      break;
  }

  for (unsigned int i=0; i<dim(); ++i)
  {
    diag_inv[i] = diag_inv[i] ? 1.0/diag_inv[i] : 1.0;	    
  }
  return instance_id;
}

void osl::stat::
SparseRegressionMultiplier::addProduct(unsigned int non_zeros,
				       const unsigned int *indices, 
				       const double *values,
				       const double *x, double *out) const
{
  assert(non_zeros <= m_dimA);
  for (unsigned int ip=0; ip<non_zeros; ++ip)
  {
    const unsigned int p = indices[ip];
    const double vp = values[ip];
    assert(p < m_dimA);
    out[p] += x[p]*vp*vp;
    for (unsigned int iq=ip+1; iq<non_zeros; ++iq)
    {
      const unsigned int q = indices[iq];
      const double vq = values[iq];
      assert(q < m_dimA);
      out[p] += x[q]*vp*vq;
      out[q] += x[p]*vp*vq;
    }
  }
}

unsigned int osl::stat::
SparseRegressionMultiplier::dim() const
{
  return m_dimA;
}

/* ------------------------------------------------------------------------- */
// CREATORS
osl::stat::
BinarySparseRegressionMultiplier::BinarySparseRegressionMultiplier(unsigned int dimA)
  : m_dimA(dimA)
{
}

osl::stat::
BinarySparseRegressionMultiplier::~BinarySparseRegressionMultiplier()
{
}

// MANIPULATORS
// ACCESSORS
void osl::stat::
BinarySparseRegressionMultiplier::newIteration() const
{
}

void osl::stat::
BinarySparseRegressionMultiplier::prod(const double *x, double *out) const
{
  newIteration();
  std::fill(&out[0], &out[m_dimA], 0.0);

  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  unsigned int non_zeros;
  while (true)
  {
    const bool go_next = getVectorX(non_zeros, &indices[0]);
    addProduct(non_zeros, &indices[0], x, out);
    if (! go_next)
      break;
  }
}

void osl::stat::
BinarySparseRegressionMultiplier::addProduct(unsigned int non_zeros,
					     const unsigned int *indices, 
					     const double *x, double *out) const
{
  assert(non_zeros <= m_dimA);
  for (unsigned int ip=0; ip<non_zeros; ++ip)
  {
    const unsigned int p = indices[ip];
    // assert(p < m_dimA);
    out[p] += x[p];
    for (unsigned int iq=ip+1; iq<non_zeros; ++iq)
    {
      const unsigned int q = indices[iq];
      // assert(q < m_dimA);
      out[p] += x[q];
      out[q] += x[p];
    }
  }
}

unsigned int osl::stat::
BinarySparseRegressionMultiplier::dim() const
{
  return m_dimA;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
