#include "osl/category/majorPiece.h"
#include "osl/category/classify.h"
#include "osl/category/classify.tcc"
#include "osl/category/categoryUtil.h"
#include "osl/record/csaString.h"
#include "osl/record/record.h"
#include "osl/record/csaRecord.h"
#include "osl/state/numEffectState.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/apply_move/applyMove.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <algorithm>
#include <fstream>

class MajorPieceTest : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(MajorPieceTest);
  CPPUNIT_TEST(testGenerate);
  CPPUNIT_TEST(testOpenAttack);
  CPPUNIT_TEST(testRookAddLanceSupport);
  CPPUNIT_TEST_SUITE_END();
public:
  void testGenerate();
  void testOpenAttack();
  void testRookAddLanceSupport();
private:
  template<class Category>
  void testGenerateOne(const char *board, const char *move);
};

CPPUNIT_TEST_SUITE_REGISTRATION(MajorPieceTest);

#include <iostream>
template<class Category>
void MajorPieceTest::
testGenerateOne(const char *board, const char *move)
{
  osl::record::csa::CsaString csa(board);
  osl::state::SimpleState initialState = csa.getInitialState();
  osl::state::NumEffectState state(initialState);
  osl::container::MoveStack history;

  osl::container::MoveLogProbVector generatedMoves;
  osl::category::Classify<osl::category::CategoryList<Category, osl::category::NullCategory> > classify;
  osl::category::CategoryEnv env(&state, osl::category::CategoryTraits::maxProb(),
				 &history);
  classify.generate(env, generatedMoves);
  if (strcmp(move, "") == 0)
  {
    if (0 != static_cast<int>(generatedMoves.size()))
    {
      std::cerr << state;
    }
    CPPUNIT_ASSERT_EQUAL(0, static_cast<int>(generatedMoves.size()));
  }
  else
  {
    if (1 != static_cast<int>(generatedMoves.size()))
    {
      std::cerr << state;
    }
    CPPUNIT_ASSERT_EQUAL(1, static_cast<int>(generatedMoves.size()));
    CPPUNIT_ASSERT_EQUAL(osl::record::csa::strToMove(move, initialState), generatedMoves[0].getMove());
  }
}

void MajorPieceTest::testGenerate()
{
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "+7868HI");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "+7868HI");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE *  *  * -KI *  * -KY\n"
     "P2-KY-GI-KI *  *  *  *  *  * \n"
     "P3-FU-FU-FU * -FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU00FU\n"
     "+\n", "+7868HI");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  *  *  *  * -KY\n"
     "P2-KY-GI *  * -KI *  *  *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU+FU-FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  *  *  *  * -KY\n"
     "P2-KY-GI-KI *  *  * +HI *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU *  * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA-HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU * -FU-GI-KE-FU-KA\n"
     "P4 *  *  * -FU * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA * +HI+KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "+6864HI");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU * -FU-GI-KE-FU-KA\n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI-FU+KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "+7868HI");
  testGenerateOne<osl::category::PrepareRookPromotion>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU * \n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5-KA *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU00FU\n"
     "P-00FU\n"
     "+\n", "");

  testGenerateOne<osl::category::RookAddPawnSupport>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI * -FU *  *  *  *  * \n"
     "P3-FU-FU-FU * -FU-GI-KE-FU-KA\n"
     "P4 *  *  * +FU * -FU-HI * -FU\n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU\n"
     "P-00FU\n"
     "+\n", "+7868HI");

  testGenerateOne<osl::category::RookAddPawnSupport>
    ("P1-OU-KE-KI *  * -KI *  * -KY\n"
     "P2-KY-GI *  *  *  *  *  *  * \n"
     "P3-FU-FU-FU-FU-FU-GI-KE-FU * \n"
     "P4 *  *  *  *  * -FU-HI * -FU\n"
     "P5-KA *  * +FU *  *  *  *  * \n"
     "P6 *  *  *  * +FU *  *  *  * \n"
     "P7+FU+FU *  * +GI+FU+FU+FU+FU\n"
     "P8 * +KA+HI * +KI * +GI+OU * \n"
     "P9+KY+KE *  *  * +KI * +KE+KY\n"
     "P+00FU\n"
     "P-00FU\n"
     "+\n", "");
}

void MajorPieceTest::testOpenAttack()
{
  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE * -KI * -HI * -KE-KY\n"
     "P2 * -OU-GI * -KI *  * -KA * \n"
     "P3 * -FU-FU-FU-FU *  *  * -FU\n"
     "P4-FU *  *  * -GI-FU-FU *  * \n"
     "P5 *  *  *  *  *  * +FU * +FU\n"
     "P6+FU * +FU * +FU *  * +GI * \n"
     "P7 * +FU+KA+FU * +FU *  * +KE\n"
     "P8 *  * +OU * +KI *  * +HI * \n"
     "P9+KY+KE+GI+KI *  *  *  * +KY\n"
     "P-00FU00FU\n"
     "+\n", "+2637GI");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE * -KI * -HI * -KE-KY\n"
     "P2 * -OU-GI * -KI *  * -KA * \n"
     "P3 * -FU-FU-FU-FU *  *  * -FU\n"
     "P4-FU *  *  *  * -FU-FU *  * \n"
     "P5 *  *  *  *  *  * +FU * +FU\n"
     "P6+FU * +FU * +FU *  * +GI * \n"
     "P7 * +FU+KA+FU * +FU *  * +KE\n"
     "P8 *  * +OU * +KI *  * +HI * \n"
     "P9+KY+KE+GI+KI *  * -GI * +KY\n"
     "P-00FU00FU\n"
     "+\n", "");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE * -KI *  *  * -KE-KY\n"
     "P2 * -OU-GI * -KI-KA * -RY * \n"
     "P3 * -FU-FU-FU-FU *  *  * -FU\n"
     "P4-FU *  *  * -GI-FU-FU *  * \n"
     "P5 *  *  *  *  *  * +FU * +FU\n"
     "P6+FU * +FU * +FU *  * +GI * \n"
     "P7 * +FU+KA+FU * +FU *  * +KE\n"
     "P8 *  * +OU * +KI *  * +RY * \n"
     "P9+KY+KE+GI+KI *  *  *  * +KY\n"
     "P-00FU00FU\n"
     "+\n", "+2637GI");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE * -KI *  *  * -KE-KY\n"
     "P2 * -OU-GI * -KI-KA * -HI * \n"
     "P3 * -FU-FU-FU-FU *  * -FU-FU\n"
     "P4-FU *  *  * -GI-FU-FU *  * \n"
     "P5 *  *  *  *  *  * +FU * +FU\n"
     "P6+FU * +FU * +FU *  * +GI * \n"
     "P7 * +FU+KA+FU * +FU *  * +KE\n"
     "P8 *  * +OU * +KI *  * +HI * \n"
     "P9+KY+KE+GI+KI *  *  *  * +KY\n"
     "P-00FU\n"
     "+\n", "");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE * -KI *  *  * -KE-KY\n"
     "P2 * -OU-GI * -KI *  * -HI * \n"
     "P3 * -FU-FU-FU-FU * -KA * -FU\n"
     "P4-FU *  *  * -GI-FU-FU *  * \n"
     "P5 *  *  *  *  *  * +FU * +FU\n"
     "P6+FU * +FU * +FU *  * +GI * \n"
     "P7 * +FU+KA+FU * +FU *  * +KE\n"
     "P8 *  * +OU * +KI *  * +HI * \n"
     "P9+KY+KE+GI+KI *  *  *  * +KY\n"
     "P-00FU00FU\n"
     "+\n", "");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE-GI-KI-OU-KI * -KE-KY\n"
     "P2 * -KA *  *  * -GI * -HI * \n"
     "P3-FU-FU-FU-FU-FU-FU * -FU-FU\n"
     "P4 *  *  *  *  *  * -FU *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU+FU+FU *  *  *  *  *  * \n"
     "P7 *  * +GI+FU+FU+FU+FU+FU+FU\n"
     "P8 * +KA+KI *  *  *  * +HI * \n"
     "P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
     "+", "+7768GI");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE-GI-KI-OU-KI * -KE-KY\n"
     "P2 *  *  *  *  * -GI * -HI * \n"
     "P3-FU-FU-FU-FU-FU-FU * -FU-FU\n"
     "P4 *  *  *  *  *  * -FU-KA * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU+FU+FU *  *  *  *  *  * \n"
     "P7 *  * +GI+FU+FU+FU+FU+FU+FU\n"
     "P8 * +UM+KI *  *  *  * +HI * \n"
     "P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
     "+", "+7768GI");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE-GI-KI-OU-KI * -KE-KY\n"
     "P2 * -HI *  *  * -KA * -GI * \n"
     "P3-FU-FU-FU-FU-FU-FU * -FU-FU\n"
     "P4 *  *  *  *  *  * -FU *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU+FU+FU *  *  *  *  *  * \n"
     "P7 *  * +GI+FU+FU+FU+FU+FU+FU\n"
     "P8 * +UM+KI *  *  *  * +HI * \n"
     "P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
     "+", "");

  testGenerateOne<osl::category::OpenAttack>
    ("P1-KY-KE-GI-KI-OU-KI * -KE-KY\n"
     "P2 * -HI *  *  * -GI * -KA * \n"
     "P3-FU-FU-FU-FU-FU *  * -FU-FU\n"
     "P4 *  *  *  *  * -FU-FU *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU+FU+FU *  *  *  *  *  * \n"
     "P7 *  * +GI+FU+FU+FU+FU+FU+FU\n"
     "P8 * +KA+KI *  *  *  * +HI * \n"
     "P9+KY+KE *  * +OU+KI+GI+KE+KY\n"
     "+", "");
}

void MajorPieceTest::testRookAddLanceSupport()
{
  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2 * -HI *  *  *  *  * -KA * \n"
     "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4 *  *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  *  * \n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
     "P8 * +KA *  *  *  *  * +HI * \n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
     "+\n", "");

  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2 * -HI *  *  *  *  * -KA * \n"
     "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4 *  *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  *  * \n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
     "P8 * +KA *  *  *  *  * +HI * \n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
     "-\n", "");

  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1 * -KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2 * -HI *  *  *  *  * -KA * \n"
     "P3-KY-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4-FU *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  * +FU\n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+KY\n"
     "P8 * +KA *  *  *  *  * +HI * \n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE * \n"
     "+\n", "+2818HI");

  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1 * -KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2 * -HI *  *  *  *  * -KA * \n"
     "P3-KY-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4-FU *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  * +FU\n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+KY\n"
     "P8 * +KA *  *  *  *  * +HI * \n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE * \n"
     "-\n", "-8292HI");

  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1 * -KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2-HI *  *  *  *  *  * -KA * \n"
     "P3-KY-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4-FU *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  * +FU\n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+KY\n"
     "P8 * +KA *  *  *  *  *  * +HI\n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE * \n"
     "+\n", "");

  testGenerateOne<osl::category::RookAddLanceSupport>
    ("P1 * -KE-GI-KI-OU-KI-GI-KE-KY\n"
     "P2-HI *  *  *  *  *  * -KA * \n"
     "P3-KY-FU-FU-FU-FU-FU-FU-FU-FU\n"
     "P4-FU *  *  *  *  *  *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6 *  *  *  *  *  *  *  * +FU\n"
     "P7+FU+FU+FU+FU+FU+FU+FU+FU+KY\n"
     "P8 * +KA *  *  *  *  *  * +HI\n"
     "P9+KY+KE+GI+KI+OU+KI+GI+KE * \n"
     "-\n", "");

}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
