#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/categoryUtil.h"
#include "osl/category/analyzer/probOfEachCategory.h"
#include "osl/category/analyzer/orderOfMove.h"
#include "osl/stat/histogram.h"
#include "osl/stat/average.h"
#include "osl/container/moveStack.h"
#include "osl/container/moveLogProbSet.h"
#include "osl/eval/pieceEval.h"
#include "osl/record/csaRecord.h"
#include "osl/record/csaString.h"
#include "osl/apply_move/applyMove.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <boost/scoped_array.hpp>
#include <string>
#include <fstream>
#include <iostream>

class StandardMoveGeneratorTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(StandardMoveGeneratorTest);
  CPPUNIT_TEST(testCover);
  CPPUNIT_TEST(testEscapeCut);
  CPPUNIT_TEST(testNotEscapeCut);
  CPPUNIT_TEST_SUITE_END();
public:
  void testCover();
  void testEscapeCut();
  void testNotEscapeCut();
};

CPPUNIT_TEST_SUITE_REGISTRATION(StandardMoveGeneratorTest);

using namespace osl;
using namespace osl::category;
using namespace osl::category::analyzer;
extern bool isShortTest;

void StandardMoveGeneratorTest::testEscapeCut()
{
  // Τä̤ǤEscapeȤ
  const char *problem = 
    "P1-KY * -GI-KI-OU-KI * -KE-KY\n"
    "P2 *  *  *  *  *  *  * -GI * \n"
    "P3-FU * +UM-FU-FU-FU * -FU-FU\n"
    "P4 *  * -FU *  *  * -FU *  * \n"
    "P5 *  *  *  *  *  *  *  *  * \n"
    "P6 * -HI *  *  *  *  *  *  * \n"
    "P7+FU *  * +FU+FU+FU+FU+FU+FU\n"
    "P8 *  * +HI *  * +OU *  *  * \n"
    "P9+KY+KE+GI+KI * +KI+GI+KE+KY\n"
    "P+00FU00KE\n"
    "P-00FU00FU00KA\n"
    "-\n";
  MoveLogProbSet allMoves;
  NumEffectState state((CsaString(problem).getInitialState()));
  MoveStack history;
  history.push(Move(Position(9,5),Position(7,3),BISHOP,
		       PTYPE_EMPTY,true,BLACK));
  CategoryEnv env(&state, 2000, &history);
  CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, allMoves);
#if 0
  for (MoveLogProbSet::const_iterator p=allMoves.begin(); p!=allMoves.end();
       ++p)
  {
    std::cerr << *p << "\n";
  }
#endif
  CPPUNIT_ASSERT_EQUAL((size_t)5, allMoves.size());
  // MoveLogProb(Move(-,KING,Position(51)->Position(52)),227)
  // MoveLogProb(Move(-,KING,Position(51)->Position(42)),227)
  // MoveLogProb(Move(-,SILVER,Position(71)->Position(62)),227)
  // MoveLogProb(Move(-,GOLD,Position(61)->Position(62)),227)
  // MoveLogProb(Drop(-,BISHOP,Position(62)),227)
}

void StandardMoveGeneratorTest::testNotEscapeCut()
{
  // ΤäƤʤ̤ǤEscapeʳƤ
  const char *problem = 
    "P1-KY * -GI-KI * -KI * -KE-KY\n"
    "P2 *  *  *  *  * -OU * -GI * \n"
    "P3-FU * +UM-FU-FU-FU * -FU-FU\n"
    "P4 *  * -FU *  *  * -FU *  * \n"
    "P5 *  *  *  *  *  *  *  *  * \n"
    "P6 * -HI *  *  *  *  *  *  * \n"
    "P7+FU *  * +FU+FU+FU+FU+FU+FU\n"
    "P8 *  * +HI *  * +OU *  *  * \n"
    "P9+KY+KE+GI+KI * +KI+GI+KE+KY\n"
    "P+00FU00KE\n"
    "P-00FU00FU00KA\n"
    "-\n";
  MoveLogProbSet allMoves;
  NumEffectState state((CsaString(problem).getInitialState()));
  MoveStack history;
  history.push(Move(Position(9,5),Position(7,3),BISHOP,
		       PTYPE_EMPTY,true,BLACK));
  CategoryEnv env(&state, 2000, &history);
  CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, allMoves);
  CPPUNIT_ASSERT_EQUAL((size_t)101, allMoves.size()); // 100+pass
}

static void testFile(std::string const& fileName)
{
  const Record rec=CsaFile(fileName).getRecord();
  NumEffectState state(rec.getInitialState());
  const vector<osl::Move> moves=rec.getMoves();
  MoveStack history;
  const size_t limit = 1800;	// ۤ800?
  CategoryEnv env(&state, limit, &history);

  stat::Histogram stat(200,10);
  stat::Average orderAverage;
  ProbOfEachCategory::Reporter reporter(std::cerr);
  for (size_t i=0; i<moves.size(); ++i)
  {
    const Move move = moves[i];

    // ˡΥƥ
    {
      MoveLogProbSet allMoves;
      CategoryListUtil::gatherAllMoves<StandardMoveGenerator>(env, allMoves);
      for (MoveLogProbSet::const_iterator p=allMoves.begin(); p!=allMoves.end(); ++p)
      {
	if (p->getMove().isPass())
	  continue;
	CPPUNIT_ASSERT(p->getMove().isValid());
	CPPUNIT_ASSERT(state.isValidMove(p->getMove()));
      }
    }
    // ΨΥƥ
    OrderOfMove::Output order;
    OrderOfMove orderAnalyzer(env, move, order);
#if 0
    const 
#endif
      MoveLogProb found = orderAnalyzer.generate(StandardMoveGenerator());
    CPPUNIT_ASSERT(found.validMove()
		   || (std::cerr << state << move << "\n", 0));
    stat.add(found.getLogProb());
    // std::cerr << "order " << order << " " << category << "\n";
    orderAverage.add(order.order);

    if ((! isShortTest) && (found.getLogProb() > 1600))
    {
      std::cerr << state << found << "\n";
      ProbOfEachCategory a(env, move, &reporter);
      a.analyze(StandardMoveGenerator());
      std::cerr << "\n";
    }
    CPPUNIT_ASSERT(found.getLogProb() < 1700);

    ApplyMoveOfTurn::doMove(state, move);
    history.push(move);
  }
  if (! isShortTest)
  {
    std::cout << fileName 
	      << " average order " << orderAverage.getAverage() << "\n";
    stat.show(std::cout);
  }
  CPPUNIT_ASSERT(orderAverage.getAverage() < 21);
}

void StandardMoveGeneratorTest::testCover()
{
  std::ifstream ifs(CSAFILES "/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName) ; i++){
    testFile(CSAFILES "/" + fileName);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
